(in-package #:lisp-sentry)


#+linux
(progn
  (declaim (inline distribution-info))
  (defun distribution-info ()
    (let* ((os-release (ignore-errors
                        (read-file-into-string #P"/etc/os-release")))
           (os-release-properties
             (when os-release
               (mapcan
                (lambda (line)
                  (when-let ((kv (uiop:split-string line :separator '(#\=)
                                                         :max 2)))
                    (setf (first kv) (make-keyword (first kv)))
                    kv))
                (uiop:split-string os-release :separator '(#\Newline))))))
      (flet ((maybe-dequote (str)
               (when str
                 (let ((length (length str)))
                   (if (and (> length 2)
                            (eql (char str 0) #\")
                            (eql (char str (1- length)) #\"))
                       (subseq str 1 (1- length))
                       str)))))
        (values
         (maybe-dequote (getf os-release-properties :id))
         (maybe-dequote (getf os-release-properties :version_id))
         (maybe-dequote (getf os-release-properties :pretty_name)))))))

(defun make-default-contexts (&key (storage-path *default-pathname-defaults*)
                                   view-names)
  "Creates list of applicable
[contexts](https://develop.sentry.dev/sdk/data-model/event-payloads/contexts)
and populates them with relevant data.

`STORAGE-PATH` is the pathname within a storage which capacity would be
reported in `DEVICE-CONTEXT`. It defaults to `*DEFAULT-PATHNAME-DEFAULTS*`.

`VIEW-NAMES` is the optional list of strings representing visible UI screens
at the current point in time, to be reported in `APP-CONTEXT`."
  (multiple-value-bind (processor-vendor processor-model arch)
      (machine-state:machine-core-info)
    (multiple-value-bind (current-charge full-charge charging-state)
        (machine-state:machine-battery)
      (multiple-value-bind (machine-vendor machine-model os-type os-version)
          (machine-state:machine-info)
        (multiple-value-bind (occupied-memory total-memory)
            (machine-state:machine-room)
          (let ((storage (ignore-errors (machine-state:storage-device
                                         storage-path))))
            (multiple-value-bind (free-storage total-storage)
                (when storage (machine-state:storage-room storage))
              (let (#+lisp-sentry-gpu
                    (display (find-if #'framebuffers:primary-p
                                      (ignore-errors
                                       (framebuffers:list-displays)))))
                (multiple-value-bind (distribution-name distribution-version
                                      distribution-pretty-name)
                    #+linux (distribution-info) #-linux nil
                  (list
                   (apply
                    #'make-device-context
                    :family processor-vendor
                    :arch arch
                    :battery-level (unless (zerop full-charge)
                                     (* 100d0 (/ current-charge full-charge)))
                    :manufacturer machine-vendor
                    :brand machine-model
                    :charging (eq charging-state :charging)
                    :memory-size total-memory
                    :free-memory (- total-memory occupied-memory)
                    :storage-size total-storage
                    :free-storage free-storage
                    :cpu-description processor-model
                    :battery-status (format nil "~:[Unknown~;~:*~@(~a~)~]"
                                            charging-state)
                    #+lisp-sentry-gpu
                    (when display
                      `(:screen-width-pixels ,(framebuffers:width display)
                        :screen-height-pixels ,(framebuffers:height display)
                        :screen-resolution
                        ,(format nil "~a⨉~a"
                                 (framebuffers:width display)
                                 (framebuffers:height display))))
                    #-lisp-sentry-gpu nil)
                   (apply
                    #'make-os-context
                    :name os-type
                    :version os-version
                    #+linux
                    `(:distribution-name ,distribution-name
                      :distribution-version ,distribution-version
                      :distribution-pretty-name ,distribution-pretty-name)
                    #-linux nil)
                   (make-runtime-context)
                   (make-app-context
                    :view-names view-names)
                   #+lisp-sentry-gpu
                   (make-gpu-context)
                   (make-feature-flag-context)))))))))))
