(in-package #:lisp-sentry)


(define-json-object (envelope)
  "[Envelopes](https://develop.sentry.dev/sdk/data-model/envelopes) are a data
format similar to HTTP form data, comprising common *Headers* and a set of
*Items* with their own headers and payloads. Envelopes are optimized for fast
parsing and human readability. They support a combination of multiple Items in
a single payload, such as:

* Submit events with large binary attachments.
* Enable communication between hops, for instance, between different
  SDKs (Native and Mobile, ReactNative and Android) and between Relays.
* Allow batching of certain Items into a single submission.
* Offline storage for deferred sending after connection issues.

See `MAKE-ENVELOPE`"
  (items nil :type list :skip t
             :documentation "the list of
[envelope items](https://develop.sentry.dev/sdk/data-model/envelope-items).")
  (processed nil :type boolean :skip t
                 :documentation "a boolean flag indicating whether this envelope
 was processed by `*PROCESS-BEFORE-SEND*` callback.")
  (dsn *dsn* :type string
             :documentation "Sentry [DSN](https://docs.sentry.io/concepts/key-terms/dsn-explainer)
string. Defaults to the value of `*DSN*`.")
  (event-id (fuuid:make-v4-string) :type string
                                   :documentation "an event ID. Autogenerated.")
  (sent-at "" :type string
              :documentation "the timestamp when the event was sent from the
client as string in RFC 3339 format. Used for clock drift correction of the
event timestamp. The time zone must be UTC. Automatically set when the
envelope is sent to Sentry."))

(defstruct (envelope-item)
  (envelope nil :type envelope))

(defgeneric dump-envelope-item-header-format (item)
  (:method (item)
    (format
     nil
     "{\"type\":~(~s~),\"content_type\":\"application/json\",\"length\":~~a}"
     (substitute #\_ #\- (string (class-name (class-of item)))))))

(defgeneric dump-envelope-item-header (item payload)
  (:method (item payload)
    (babel:string-to-octets
     (format nil (dump-envelope-item-header-format item) (length payload)))))

(defgeneric dump-envelope-item-payload (item))

(defun dump-envelope-item (item)
  (flex:with-output-to-sequence (stream)
    (let ((payload (dump-envelope-item-payload item)))
      (write-sequence (dump-envelope-item-header item payload) stream)
      (write-byte (char-code #\Newline) stream)
      (write-sequence payload stream))))

(defun dump-envelope* (envelope)
  (flex:with-output-to-sequence (stream)
    (write-sequence (babel:string-to-octets (dump-envelope envelope)) stream)
    (write-byte (char-code #\Newline) stream)
    (dolist (item (reverse (envelope-items envelope)))
      (write-sequence (dump-envelope-item item) stream)
      (write-byte (char-code #\Newline) stream))))

(declaim
 (type (or symbol (function (envelope) (or envelope envelope-item null)))
       *process-before-send*))
(defvar *process-before-send* #'identity
  "Designator of function of one argument. The function receives the `ENVELOPE`
object right before it is sent to Sentry. **NOTE**: the return value of the
function is passed instead of original envelope to `CAPTURE`. If the function
returns `NIL`, nothing gets sent.

Defaults to `IDENTITY`.")

(defgeneric capture (object)
  (:documentation "The reporting of an event is called capturing. When an event
 is captured, it’s sent to Sentry.

`OBJECT` should be `ENVELOPE` object or one of envelope items (currently
supported items are `ATTACHMENT`, `EVENT` and `USER-REPORT`), in which case
the envelope where the item belongs is captured.")
  (:method ((object envelope))
    (if (envelope-processed object)
        (progn
          (setf (envelope-sent-at object)
                (local-time:to-rfc3339-timestring (local-time:now)))
          (send-payload "envelope" :content-type "application/x-sentry-envelope"
                                   :content (dump-envelope* object)))
        (when-let (processed (funcall *process-before-send* object))
          (setf (envelope-processed object) t)
          (capture processed))))
  (:method ((object envelope-item))
    (capture (envelope-item-envelope object))))
