(in-package #:lisp-sentry)


(defmacro define-json-object ((name &key ctor include) docstring &rest slots)
  (with-gensyms (x)
    (let* ((slot-names (mapcar #'first slots))
           (slot-defaults (mapcar #'second slots))
           (slot-types (mapcar (lambda (s) (getf s :type)) slots))
           (slot-optional (mapcar (lambda (s) (getf s :optional)) slots))
           (slot-raw (mapcar (lambda (s) (getf s :raw)) slots))
           (slot-skip (mapcar (lambda (s) (getf s :skip)) slots))
           (slot-docs (mapcar (lambda (s) (getf s :documentation "")) slots))
           (slot-types*  (mapcar (lambda (type o) (if o `(or ,type null) type))
                                 slot-types slot-optional))
           (slot-format (mapcar (lambda (s) (getf s :format "~s")) slots))
           (slot-processors (mapcar (lambda (s) (getf s :processor ''identity)) slots))
           (slot-accessors (mapcar (lambda (n p o)
                                     (if o
                                         `(when-let (,n (,(symbolicate name :- n) ,x))
                                            (funcall ,p ,n))
                                         `(funcall ,p (,(symbolicate name :- n) ,x))))
                                   slot-names slot-processors slot-optional))
           (last-slot (car (last slot-names)))
           (slot-format* (mapcar (lambda (n o f r s)
                                   (if s
                                       "~*"
                                       (format nil "~a~:[~s:~;~*~]~a~a~a~a"
                                               (if o "~:[~;" "")
                                               r
                                               (nsubstitute #\_ #\- (string-downcase n))
                                               (if o "~:*" "")
                                               f
                                               (if (eq n last-slot) "" ",")
                                               (if o "~]" ""))))
                                 slot-names slot-optional slot-format
                                 slot-raw slot-skip)))
      (when (car (last slot-optional))
        (error "Last slot is supposed to be non-optional"))
      `(progn
         (defstruct (,name
                     ,@(when ctor `((:constructor ,ctor)))
                     ,@(when include `((:include ,include))))
           ,docstring
           ,@(mapcar (lambda (n d type)
                       `(,n ,d :type ,type))
              slot-names slot-defaults slot-types*))
         (setf (documentation #',(if ctor ctor (symbolicate :make- name))
                              'function)
               ,(format nil
                        "Creates `~a`.~%~%~{* `~a` is ~:[~;optional ~]~a~^~%~}"
                        name
                        (mapcan #'list slot-names slot-optional slot-docs)))
         (declaim (ftype (function (,name) (values string &optional))
                         ,(symbolicate :dump- name)))
         (defun ,(symbolicate :dump- name) (,x)
           (format nil ,(apply #'concatenate 'string
                               "{" (append slot-format* '("}")))
                   ,@slot-accessors))))))

(eval-when (:compile-toplevel :load-toplevel)
  (defun system-version (system)
    (when-let (asdf-system (asdf:find-system system nil))
      (asdf:component-version asdf-system)))

  (declaim (ftype (function (&optional list) (values list &optional))
                  get-modules))
  (defun get-modules (&optional list)
    "Returns plist of already loaded ASDF systems and their versions.

If `LIST` argument is given, the function returns those systems instead."
    (loop :for system-name :of-type string :in (asdf:already-loaded-systems)
          :for system := (asdf:find-system system-name nil)
          :unless (and list (not (member system-name list :test #'string=)))
          :when system
          :nconc (list system-name
                       (or (asdf:component-version system) "#<UNKNOWN>"))))

  (defun package-files (name)
    (labels ((component-pathnames (obj)
               (when (typep obj 'asdf:module)
                 (let ((components (asdf:module-components obj)))
                   (nconc
                    (mapcar
                     (lambda (component)
                       (when (typep component 'asdf:cl-source-file)
                         (asdf:component-pathname component)))
                     components)
                    (mapcan #'component-pathnames components))))))
      (remove nil (component-pathnames (asdf:find-system name)) :test #'eq))))

(defun escape-nonprintable (string)
  (ppcre:regex-replace-all
   "[\\x00-\\x1F\\x7F]"
   string
   (lambda (match)
     (let ((char (char-code (char match 0))))
       (case char
         (7  "\\a")
         (8  "\\b")
         (9  "\\t")
         (10 "\\n")
         (11 "\\v")
         (12 "\\f")
         (13 "\\r")
         (otherwise (format nil "\\x~2,'0x" char)))))
   :simple-calls t))

(defmacro ignore-errors* (&rest forms)
  `(handler-case (progn ,@forms)
     (error (condition)
       (declare (ignore condition))
       (values nil))))

(defun reverse-plist (plist)
  (loop :with result := nil
        :for (k v) :on plist :by #'cddr
        :do (setf result (list* k v result))
        :finally (return result)))
