(in-package #:lisp-sentry)


;; get-thread-id 

(defun get-thread-id (thread)
  (sb-kernel:get-lisp-obj-address (bt2:thread-native-thread thread)))

;; make-source 

(defstruct (source* (:include source))
  (start-rows (make-hash-table :test #'eql) :type hash-table)
  (start-cols (make-hash-table :test #'eql) :type hash-table)
  (end-rows (make-hash-table :test #'eql) :type hash-table)
  (end-cols (make-hash-table :test #'eql) :type hash-table))

(declaim (ftype (function (array-index array-index) (values fixnum &optional))
                location->fixnum)
         (inline location->fixnum))
(defun location->fixnum (tlf-offset form-number)
  (the fixnum
       (logior (logand form-number #.(1- (ash 1 30))) (ash tlf-offset 30))))

(defun make-sharpdot-reader (orig-sharpdot-reader)
  (lambda (s c n)
    (ignore-errors (funcall orig-sharpdot-reader s c n))))

(defun make-source-recorder (fn source-map)
  (lambda (stream char)
    (let ((start (1- (file-position stream)))
          (values (multiple-value-list (funcall fn stream char)))
          (end (file-position stream)))
      (when values
        (destructuring-bind (&optional existing-start &rest existing-end)
            (car (gethash (car values) source-map))
          (unless (and existing-start existing-end
                       (<= start existing-start end)
                       (<= start existing-end end))
            (push (cons start end) (gethash (car values) source-map)))))
      (values-list values))))

(defun make-source-recording-readtable (readtable source-map)
  (declare (type readtable readtable) (type hash-table source-map))
  (flet ((install-special-sharpdot-reader (rt)
           (let ((fun (ignore-errors
                       (get-dispatch-macro-character #\# #\. rt))))
             (when fun
               (let ((wrapper (make-sharpdot-reader fun)))
                 (set-dispatch-macro-character #\# #\. wrapper rt)))))
         (install-wrappers (rt)
           (dotimes (code 128)
             (let ((char (code-char code)))
               (multiple-value-bind (fun nt) (get-macro-character char rt)
                 (when fun
                   (let ((wrapper (make-source-recorder fun source-map)))
                     (set-macro-character char wrapper nt rt))))))))
    (let ((rt (copy-readtable readtable)))
      (install-special-sharpdot-reader rt)
      (install-wrappers rt)
      rt)))

(defgeneric sexp-in-bounds-p (sexp i)
  (:method ((list list) i)
    (< i (loop :for e :on list :count t)))
  (:method ((sexp t) i) nil))

(defgeneric sexp-ref (sexp i)
  (:method ((s list) i) (elt s i)))

(defmethod sexp-in-bounds-p ((s sb-impl::comma) i)
  (sexp-in-bounds-p (sb-impl::comma-expr s) i))

(defmethod sexp-ref ((s sb-impl::comma) i)
  (sexp-ref (sb-impl::comma-expr s) i))

(defun source-path-source-position (path form source-map)
  (let ((forms (loop :for i :in path
                     :for f := form :then (if (sexp-in-bounds-p f i)
                                              (sexp-ref f i))
                     :collect f)))
    (loop :for form :in (nreverse forms)
          :for ((start . end) . rest) := (gethash form source-map)
          :when (and start end (not rest))
            :return (return (values start end)))))

(declaim (ftype (function ((simple-array array-index) array-index)
                          (values array-index &optional))
                binary-search))
(defun binary-search (data value)
  (declare (optimize (safety 0)))
  (loop :with l :of-type array-index := 0
        :with r :of-type array-index := (1+ (length data))
        :while (< l r)
        :for i :of-type array-index := (floor (the array-index (+ l r)) 2)
        :for a := (aref data i)
        :do (if (> a value)
                (setf r i)
                (setf l (the array-index (1+ i))))
        :finally (return r)))

(defun extract-package (line)
  (declare (type string line))
  (let ((name (cadr (read-from-string line))))
    (find-package name)))

(defun guess-package (stream)
  (let* ((pkg *package*))
    (loop :for read-line := (read-line stream nil nil)
          :for line
            := (and read-line
                    (string-trim
                     '(#\Space #\Tab #\Linefeed #\Page #\Return #\Rubout)
                     read-line))
          :do (when (not line) (return))
              (when (or (uiop:string-prefix-p "(in-package " line)
                        (uiop:string-prefix-p "(cl:in-package " line))
                (let ((p (extract-package line)))
                  (when p (setf pkg p)))
                (return)))
    (file-position stream 0)
    pkg))

(defun make-source (contents)
  (let* ((lines (uiop:split-string contents :separator '(#\Newline)))
         (line-offsets* (nreverse (reduce
                                   (lambda (offsets line)
                                     (push (+ 1 (first offsets) (length line))
                                           offsets))
                                   lines
                                   :initial-value (list 0))))
         (line-offsets (make-array (length line-offsets*)
                                   :element-type 'array-index
                                   :initial-contents line-offsets*))
         (start-rows (make-hash-table :test #'eql))
         (start-cols (make-hash-table :test #'eql))
         (end-rows (make-hash-table :test #'eql))
         (end-cols (make-hash-table :test #'eql))
         (*features* (append *features*
                             (symbol-value
                              (find-symbol "+INTERNAL-FEATURES+" 'sb-impl)))))
    (declare (dynamic-extent line-offsets* line-offsets))
    (proclaim '(sb-ext:muffle-conditions warning))
    (with-input-from-string (stream contents)
      (loop :with source-map :of-type hash-table := (make-hash-table :test #'eq)
            :with *readtable*
               := (make-source-recording-readtable *readtable* source-map)
            :with *package* := (guess-package stream)
            :with *read-suppress* := nil
            :for tlf-number :of-type array-index :from 0
            :for tlf := (ignore-errors (read stream nil 'eof))
            :for path-table := (sb-di::form-number-translations tlf 0)
            :until (eq tlf 'eof)
            :do (loop :for form-number :of-type array-index :from 0
                      :for index := (location->fixnum tlf-number form-number)
                      :for path* :across path-table
                      :for path := (reverse (cdr path*))
                      :for (start end) := (multiple-value-list
                                           (source-path-source-position
                                            path tlf source-map))
                      :when start
                      :do (let* ((start-line
                                   (binary-search line-offsets start))
                                 (start-line-offset
                                   (aref line-offsets
                                         (min (1- start-line)
                                              (1- (length line-offsets)))))
                                 (end-line
                                   (binary-search line-offsets end))
                                 (end-line-offset
                                   (aref line-offsets
                                         (min (1- end-line)
                                              (1- (length line-offsets))))))
                            (setf (gethash index start-rows)
                                  start-line
                                  (gethash index start-cols)
                                  (- start start-line-offset -1)
                                  (gethash index end-rows)
                                  end-line
                                  (gethash index end-cols)
                                  (- end end-line-offset -1))))))
    (proclaim '(sb-ext:unmuffle-conditions warning))
    (make-source*
     :lines (make-array (1- (length line-offsets))
                        :element-type 'string
                        :initial-contents lines)
     :start-rows start-rows
     :start-cols start-cols
     :end-rows end-rows
     :end-cols end-cols)))

;; get-stack-trace 

(defun code-location-has-debug-block-info-p (code-location)
  (handler-case
      (progn (sb-di:code-location-debug-block code-location)
             t)
    (sb-di:no-debug-blocks  () nil)))

(defun maybe-fold-path (function)
  (let ((folded (when-let* ((listp (listp function))
                            (plistp (evenp (length function)))
                            (path (getf function :in))
                            (stringp (stringp path)))
                  (let ((result (copy-list function)))
                    (setf (getf result :in) (file-namestring path))
                    result))))
    (if folded folded function)))

(declaim (inline system-source-p))
(defun system-source-p (file)
  (when file
    (or
     (uiop:string-prefix-p
      #.(namestring (translate-logical-pathname "SYS:SRC;"))
      (namestring file))
     (uiop:string-prefix-p
      #.(namestring (translate-logical-pathname "SYS:CONTRIB;"))
      (namestring file)))))

(defun get-stack-trace ()
  (let (frames)
    (sb-debug:map-backtrace
     (lambda (frame)
       (let* ((location (sb-di:frame-code-location frame))
              (start-location
                (when (code-location-has-debug-block-info-p location)
                  (sb-debug::maybe-block-start-location location)))
              (tlf-offset (when start-location
                            (sb-di:code-location-toplevel-form-offset
                             start-location)))
              (form-number (when start-location
                             (sb-di:code-location-form-number
                              start-location)))
              (form-index (when start-location
                            (location->fixnum tlf-offset form-number)))
              (dsource (ignore-errors
                        (sb-di:code-location-debug-source location)))
              (source-file (when dsource
                             (sb-di:debug-source-namestring dsource)))
              (source-file* (when source-file
                              (translate-logical-pathname source-file)))
              (source (when source-file*
                        (gethash (namestring source-file*) *sources*)))
              (start-line (when source
                            (gethash form-index (source*-start-rows source))))
              (end-line (when source
                          (gethash form-index (source*-end-rows source))))
              (dfun (sb-di:frame-debug-fun frame))
              (fun (sb-di:debug-fun-fun dfun))
              (fun-symbol (ignore-errors (sb-impl::%fun-name fun)))
              (package (typecase fun-symbol
                         (null nil)
                         (symbol (symbol-package fun-symbol))
                         ((cons (eql setf) (cons symbol))
                          (symbol-package (second fun-symbol)))
                         ((cons (eql flet) (cons symbol cons))
                          (symbol-package (second fun-symbol)))
                         ((cons (eql lambda) (cons (or null list)
                                              (cons (eql :in) (cons symbol))))
                          (symbol-package (fourth fun-symbol)))
                         ((cons (eql labels) (cons symbol
                                                   (cons (eql :in)
                                                         (cons symbol))))
                          (symbol-package (fourth fun-symbol)))
                         ((cons (eql sb-pcl::fast-method) (cons symbol cons))
                          (symbol-package (second fun-symbol)))
                         ((cons (eql sb-pcl::slow-method) (cons symbol cons))
                          (symbol-package (second fun-symbol)))))
              (ptr (sb-sys:sap-int (sb-di::frame-pointer frame)))
              (vars (sb-di::debug-fun-debug-vars dfun)))
         (push
          (guess-source-link
           (make-stack-trace-frame
            :filename (abs-path->filename source-file*)
            :abs-path source-file*
            :function (format nil "~a"
                              (if fun-symbol
                                  (maybe-fold-path fun-symbol)
                                  (sb-debug::frame-call frame)))
            :instruction-addr ptr
            :package (when package
                       (package-name package))
            :lineno start-line
            :colno (when source
                     (gethash form-index (source*-start-cols source)))
            :context-line (when start-line
                            (aref (source-lines source) (1- start-line)))
            :pre-context (when start-line
                           (coerce
                            (subseq (source-lines source)
                                    (max 0 (- start-line 6))
                                    (1- start-line))
                            'list))
            :post-context (when end-line
                            (coerce
                             (subseq (source-lines source)
                                     start-line
                                     (min (+ 5 start-line)
                                          (length (source-lines source))))
                             'list))
            :vars
            (when vars
              (loop
                :for var :across vars
                :for var-name := (sb-di:debug-var-symbol var)
                :with seen := nil
                :nconc
                   (unless (or (find var-name seen)
                               (length= 0 (string var-name)))
                     (push var-name seen)
                     `(,(string var-name)
                       ,(if (eq (sb-di:debug-var-validity var location) :valid)
                            (pretty-print (sb-di:debug-var-value var frame))
                            (princ-to-string (sb-int:make-unprintable-object
                                              "UNAVAILABLE VARIABLE")))))))))
          frames))))
    (when-let
        (self-frame
         (position-if
          (lambda (frame)
            (and
             (string=
              "(FLET THREAD-STACK-TRACE-PROBE IN GET-THREAD-STACK-TRACE)"
              (stack-trace-frame-function frame))
             (stack-trace-frame-package frame)
             (string= "LISP-SENTRY" (stack-trace-frame-package frame))))
          frames))
      (setf frames (subseq frames 0 self-frame)))
    (loop :with in-app := nil
          :for f :of-type stack-trace-frame :in frames
          :do (unless in-app
                (when (or
                       ;; use *project-package*
                       (and *project-package*
                            (string-equal (package-name *project-package*)
                                          (stack-trace-frame-package f)))
                       ;; use *project-root*
                       (and *project-root*
                            (uiop:string-prefix-p
                             (namestring *project-root*)
                             (namestring
                              (or (stack-trace-frame-abs-path f) ""))))
                       ;; if not given, use simple system-source-p heuristic
                       (and
                        (not *project-package*)
                        (not (system-source-p (stack-trace-frame-abs-path f)))
                        (not (uiop:string-prefix-p
                              "foreign function"
                              (stack-trace-frame-function f)))))
                  (setf in-app t)))
              (setf (stack-trace-frame-in-app f) in-app))
    frames))
