(in-package #:lisp-sentry)


(defvar *breadcrumbs* nil)

(define-json-object (breadcrumb :ctor %make-breadcrumb)
  "Sentry uses [breadcrumbs](https://develop.sentry.dev/sdk/data-model/event-payloads/breadcrumbs)
to create a trail of events that happened prior to an issue. These events are
very similar to traditional logs but can record more rich structured data.

See `MAKE-BREADCRUMB`, `ADD-BREADCRUMBS`"
  (type "default" :type string)
  (category nil :type string :optional t)
  (message nil :type string :optional t)
  (data nil :type list :optional t :format "~{\"~(~a~)\":~s~^,~}")
  (level :info :type level :format "\"~(~a~)\"")
  (origin nil :type string :optional t)
  (timestamp (local-time:to-rfc3339-timestring (local-time:now)) :type string))

(define-json-object (breadcrumbs-interface)
  "Breadcrumbs event interface.

See `MAKE-BREADCRUMBS-INTERFACE`, `BREADCRUMB`"
  (values nil :type list :format "[~{~a~^,~}]"
              :processor (lambda (breadcrumbs)
                           (nreverse (mapcar #'dump-breadcrumb breadcrumbs)))
              :documentation "a list of `BREADCRUMB` objects."))

(declaim (inline make-breadcrumb-data))
(defun make-breadcrumb-data (&rest data &key &allow-other-keys)
  "Prepares the arbitrary keyword arguments to be used as the `DATA` for
`MAKE-BREADCRUMB`."
  data)

(defun make-breadcrumb (&key
                        (type "default")
                        category
                        message
                        (data (make-breadcrumb-data))
                        (level :info)
                        origin
                        (timestamp nil timestamp-supplied-p))
  "Creates a `BREADCRUMB`.

* `TYPE` is the type of breadcrumb. By default, all breadcrumbs are recorded
as `default`, which makes them appear as a Debug entry, but Sentry provides
other types that influence how the breadcrumbs are rendered. For more
information, see the [description of recognized breadcrumb
types](https://develop.sentry.dev/sdk/data-model/event-payloads/breadcrumbs/#breadcrumb-types).
* `CATEGORY` is an optional dotted string indicating what the crumb is or from
where it comes. Typically it is a module name or a descriptive string. For
instance, `ui.click` could be used to indicate that a click happened in the UI
or `flask` could be used to indicate that the event originated in the Flask
framework.
* `MESSAGE` is optional human-readable message for the breadcrumb.
* `DATA` is optional arbitrary data associated with this breadcrumb. See
`MAKE-BREADCRUMB-DATA`.
* `LEVEL` is optional severity level of the breadcrumb. Allowed values are,
from highest to lowest: `:FATAL`, `:ERROR`, `:WARNING`, `:INFO`, and
`:DEBUG`. Levels are used in the UI to emphasize and deemphasize the
crumb. The default is `:INFO`.
* `ORIGIN` is a string representing the origin of the breadcrumb. This is
typically used to identify the source of the breadcrumb. For example hybrid
SDKs can identify native breadcrumbs from JS or Flutter.
* `TIMESTAMP` is a timestamp representing when the breadcrumb
occurred. Breadcrumbs are most useful when they include a timestamp, as it
creates a timeline leading up to an event \"exception\"/\"error\". This
defaults to current time.

See `ADD-BREADCRUMBS`"
  (let ((breadcrumb (apply #'%make-breadcrumb
                           :type type
                           :category category
                           :message message
                           :data data
                           :level level
                           :origin origin
                           (when timestamp-supplied-p
                             `(:timestamp ,timestamp)))))
    (push breadcrumb *breadcrumbs*)
    breadcrumb))

(defun add-breadcrumbs (event)
  "Adds all breadcrumbs previously created with `MAKE-BREADCRUMB` to the
`EVENT`.

**NOTE** that breadcrumbs are automatically added by `MAKE-MESSAGE-EVENT` and
`MAKE-CONDITION-EVENT`."
  (push
   (make-breadcrumbs-interface :values *breadcrumbs*)
   (event-interfaces event))
  (setf *breadcrumbs* nil)
  event)

(defmethod dump-event-interface ((interface breadcrumbs-interface))
  `("breadcrumbs"
    ,(dump-breadcrumbs-interface interface)))
