(in-package #:lisp-sentry)


(declaim (type boolean *exception-handled*))
(defvar *exception-handled* nil
  "A global boolean flag indicating whether the currently processed condition
was handled. Defaults to `NIL`.")

(define-json-object (exception-mechanism)
  "The [exception mechanism](https://develop.sentry.dev/sdk/data-model/event-payloads/exception/#exception-mechanism)
is an optional field residing in the `EXCEPTION-INTERFACE`. It carries
additional information about the way the exception was created on the target
system. This includes general exception values obtained from the operating
system or runtime APIs, as well as mechanism-specific values.

See `MAKE-EXCEPTION-MECHANISM`, `ADD-EXCEPTION`"
  (type "condition"
        :type string
        :documentation "unique identifier of this mechanism determining
rendering and processing of the mechanism data. Defaults to \"condition\".")
  (description
   nil
   :type string :optional t
   :documentation "human-readable description of the error mechanism and a
possible hint on how to solve this error.")
  (help-link nil :type string :optional t
                 :documentation "fully qualified URL to an online help resource,
possibly interpolated with error parameters.")
  (synthetic
   nil
   :type boolean :format "~:[false~;true~]"
   :documentation "an optional flag indicating that this error is synthetic.
Synthetic errors are errors that carry little meaning by themselves. This may
be because they are created at a central place (like a crash handler), and are
all called the same: `Error`, `Segfault` etc.  When the flag is set, Sentry
will then try to use other information (top in-app frame function) rather than
the exception type and value in the UI for the primary event display.
Furthermore, if this flag is set, Sentry will ignore the exception type when
grouping the exception into issues.
This flag should be set for all \"segfaults\" for instance as every single
error group would look very similar otherwise. Also, errors the client creates
to add a stack trace to events that don't have one themselves should be marked
as `synthetic` (This happens, for example, when users set `attachStackTrace:
true` and capture a string message via `captureException` or
`captureMessage`).")
  (exception-id nil :type fixnum :optional t
                    :documentation "numeric value providing an ID for the
exception relative to this specific event.")
  (parent-id nil :type fixnum :optional t
                 :documentation "numeric value pointing at the `exception-id`
that is the parent of this exception. ")
  (is-exception-group nil :type boolean :format "~:[false~;true~]"
                          :documentation "flag indicating that this exception is
part of an exception group type specific to the platform or language.")
  (meta nil :type t :optional t :format "~a"
            :documentation "information from the operating system or runtime on
the exception mechanism (see
[meta information](https://develop.sentry.dev/sdk/data-model/event-payloads/exception/#meta-information)).
**NOTE** that this should be printed with `~A` to valid JSON string.")
  (data nil :type t :optional t :format "~a"
            :documentation "arbitrary extra data that might help the user
understand the error thrown by this mechanism. **NOTE** that this should be
printed with `~A` to valid JSON string.")
  (handled *exception-handled*
           :type boolean :format "~:[false~;true~]"
           :documentation "flag indicating whether the user has handled the
exception (for example, via `HANDLER-CASE`).
Defaults to `*EXCEPTION-HANDLED*`."))

(define-json-object (exception-interface)
  "The [Exception Interface](https://develop.sentry.dev/sdk/data-model/event-payloads/exception)
specifies an exception or error that occurred in a program.

See `MAKE-EXCEPTION-INTERFACE`"
  (type "" :type string :documentation "the type of exception.")
  (value "" :type string :documentation "the value of exception (a string).")
  (module "" :type string :documentation "the optional module, or package which
the exception type lives in.")
  (thread-id 0 :type array-length :format "\"0x~x\"" :documentation "an optional
value which refers to a thread in the `THREADS-INTERFACE`.")
  (mechanism nil :type exception-mechanism
                 :optional t
                 :processor 'dump-exception-mechanism
                 :format "~a"
                 :documentation "object describing the mechanism that created
this exception. See `EXCEPTION-MECHANISM`.")
  (stacktrace nil :type list :format "~a" :processor 'dump-stack-trace
                  :documentation "an optional stack trace object, a list of
`STACK-TRACE-FRAME` objects."))

(defun add-exception (event exception stack-trace thread)
  "Adds `EXCEPTION-INTERFACE` with info on given `EXCEPTION` happened in
`THREAD` with `STACK-TRACE` to given `EVENT`.

See `GET-STACK-TRACE`"
  (push
   (make-exception-interface
    :type (string (type-of exception))
    :value (princ-to-string exception)
    :module (package-name (symbol-package (type-of exception)))
    :thread-id (get-thread-id thread)
    :mechanism (make-exception-mechanism)
    :stacktrace stack-trace)
   (event-interfaces event))
  event)

(defmethod dump-event-interface ((interface exception-interface))
  `("exception"
    ;; NOTE: for now, only single exception is supported
    ,(format nil "{\"values\":[~a]}" (dump-exception-interface interface))))
