(in-package #:lisp-sentry)


(define-json-object (stack-trace-frame)
  "A [stack trace](https://develop.sentry.dev/sdk/data-model/event-payloads/stacktrace)
contains a list of frames, each with various bits (most optional) describing
the context of that frame.

See `MAKE-STACK-TRACE-FRAME`, `GET-STACK-TRACE`"
  (filename nil :type string :optional t
                :documentation "path to the source file relative to the project
root directory (see `*PROJECT-ROOT*`).")
  (lineno nil :type fixnum :optional t
              :documentation "line number of the call, starting at 1.")
  (colno nil :type fixnum :optional t
             :documentation "column number of the call, starting at 1.")
  (abs-path nil :type pathname :optional t :processor 'namestring
                :documentation "absolute path to the source file.")
  (context-line nil :type string :optional t
                    :documentation "source code in filename at `LINENO`.")
  (pre-context nil :type list :optional t :format "[~{~s~^,~}]"
                   :documentation "list of source code lines before
`CONTEXT-LINE` (in order).")
  (post-context nil :type list :optional t :format "[~{~s~^,~}]"
                    :documentation "list of source code lines after
`CONTEXT-LINE` (in order).")
  (source-link nil :type string :optional t
                   :documentation "URL representing the source code, e.g.
commit-specific GitHub raw source link.")
  (in-app nil :type boolean :format "~:[false~;true~]"
              :documentation "the flag indicating whether this frame is related
to the execution of the relevant code in this stack trace. For example, the
frames that might power the framework’s web server of your app are probably not
relevant. However, calls to the framework’s library once you start handling code
likely are relevant.")
  (vars nil :type list :optional t :format "{~{~s:~s~^,~}}"
            :documentation "mapping of variables which were available within
this frame (usually context-locals).")
  (instruction-addr 0 :type non-negative-fixnum :format "\"0x~x\""
                      :documentation "instruction address for symbolication.")
  (package nil :type string :optional t
               :documentation "\"package\" the frame was contained in.")
  (function "" :type string
               :documentation "the name of the function being called."))

(defvar *project-root* nil
  "Absolute path to the project directory. Defaults to `NIL`.")
(defvar *project-package* nil
  "Name of the project package. Used to determine whether the stack frame is
related to the app. Defaults to `NIL`.")

(defun abs-path->filename (abs-path)
  (when abs-path
    (if *project-root*
        (let ((namestring (enough-namestring abs-path *project-root*)))
          (if (uiop:relative-pathname-p namestring)
              namestring
              (file-namestring abs-path)))
        (file-namestring abs-path))))

(defun github-project-root (github-url)
  (ppcre:register-groups-bind (user-1 user-2 repo)
      ("(?:https?://|//)?(?:github.com/([\\w-]+)|([\\w-]+).github.io)/([\\w-]+)"
       github-url)
    (format nil "https://github.com/~a/~a" (or user-1 user-2) repo)))

(defun gitlab-project-root (gitlab-url)
  (ppcre:register-groups-bind (user-1 user-2 repo)
      ("(?:https?://|//)?(?:gitlab.com/([\\w-]+)|([\\w-]+).gitlab.io)/([\\w-]+)"
       gitlab-url)
    (format nil "https://gitlab.com/~a/~a" (or user-1 user-2) repo)))

(defun guess-source-link (frame)
  (when-let* ((lineno (stack-trace-frame-lineno frame))
              (abs-path (stack-trace-frame-abs-path frame))
              (package (stack-trace-frame-package frame))
              ;; NOTE: assuming system name coincides with package name
              (system (ignore-errors (asdf:find-system package)))
              (homepage (asdf:system-homepage system))
              (source-dir (asdf:system-source-directory system)))
    (setf (stack-trace-frame-source-link frame)
          (cond ((search "github" homepage)
                 (format nil "~a/blob/HEAD/~a~@[#L~a~]"
                         (github-project-root homepage)
                         (enough-namestring abs-path source-dir)
                         lineno))
                ((search "gitlab" homepage)
                 (format nil "~a/blob/HEAD/~a~@[#L~a~]"
                         (github-project-root homepage)
                         (enough-namestring abs-path source-dir)
                         lineno))
                (t
                 nil))))
  frame)

(declaim
 (inline get-stack-trace)
 (ftype (function () (values list &optional)) get-stack-trace))

(setf (documentation 'get-stack-trace 'function)
      "Returns a list of `STACK-TRACE-FRAME` objects corresponding to current
stack trace, sorted from oldest to newest.")

(defun dump-stack-trace (frames)
  (format nil "{\"frames\":[~{~a~^,~}]}"
          (mapcar #'dump-stack-trace-frame frames)))
