(in-package #:lisp-sentry)


(define-json-object (thread)
  "An object containing information on thread within `THREADS-INTERFACE`.

See `MAKE-THREAD`, `GET-THREADS`, `ADD-THREADS`"
  (id 0 :type array-length :format "\"0x~x\""
        :documentation "the ID of the thread. Typically a number or numeric
 string. Needs to be unique among the threads. An exception can set the
`THREAD-ID` attribute to cross-reference this thread.")
  (crashed nil :type boolean :format "~:[false~;true~]"
               :documentation "a flag indicating whether the thread was the
cause for the event being sent, e.g. due to a crash. Defaults to `NIL`.")
  (current nil :type boolean :format "~:[false~;true~]"
               :documentation "a flag indicating whether the thread was in the
foreground. A thread is in foreground when it's executing. Defaults to `NIL`.")
  (stacktrace nil :type list :optional t :format "~a"
                  :processor 'dump-stack-trace
                  :documentation "stack trace. List of `STACK-TRACE-FRAME`
objects.")
  (name "" :type string :documentation "the thread name."))

(define-json-object (threads-interface)
  "The [Threads Interface](https://develop.sentry.dev/sdk/data-model/event-payloads/threads)
specifies threads that were running at the time an event happened. These
threads can also contain stack traces.

See `MAKE-THREADS-INTERFACE`, `GET-THREADS`"
  (values nil :type list :format "[~{~a~^,~}]"
              :processor (lambda (threads) (mapcar #'dump-thread threads))
              :documentation "the list of `THREAD` objects."))

(defun get-thread-stack-trace (thread)
  (let ((cv (bt2:make-condition-variable))
        (lock (bt2:make-lock))
        stack-trace
        done)
    (flet ((thread-stack-trace-probe ()
             (#+sbcl sb-sys:with-interrupts #-sbcl progn
              (bt2:with-lock-held (lock)
                (setf stack-trace (get-stack-trace)
                      done t)
                (bt2:condition-notify cv)))))
      (bt2:with-lock-held (lock)
        (bt2:interrupt-thread thread #'thread-stack-trace-probe)
        (tagbody
         wait
           (bt2:condition-wait cv lock)
           (unless done
             (go wait)))
        stack-trace))))

(defun get-threads (crashed)
  "Gets the list of `THREAD` objects corresponding to the list of all threads
returned by `(BT2:ALL-THREADS)`. `CRASHED` is the thread that has crashed, or
`NIL`."
  (loop :with current := (bt2:current-thread)
        :for thread :in (bt2:all-threads)
        :for currentp := (eq thread current)
        :collect (make-thread :id (get-thread-id thread)
                              :crashed (eq thread crashed)
                              :current currentp
                              :name (bt2:thread-name thread)
                              :stacktrace (unless currentp
                                            (get-thread-stack-trace thread)))))

(defun add-threads (event &key (crashed (bt2:current-thread)))
  "Adds `THREADS-INTERFACE` to the given `EVENT`. `CRASHED` is the crashed
thread, defaulting to the current thread."
  (push
   (make-threads-interface :values (get-threads crashed))
   (event-interfaces event))
  event)

(defmethod dump-event-interface ((interface threads-interface))
  `("threads"
    ,(dump-threads-interface interface)))
