(in-package #:lisp-sentry)


(deftype level () '(member :fatal :error :warning :info :debug))

(declaim (type string *environment*))
(defvar *environment* "production"
  "The environment name, such as `production` or `staging`. Defaults to
`production`.")
(declaim (type (or string null) *server-name* *release* *dist*))
(defvar *server-name* nil
  "Name of the device, typically a hostname. Defaults to `NIL`.")
(defvar *release* nil
  "The release version of the application. Release versions must be unique
 across all projects in your organization. This value can be the git SHA for
the given project, or a product identifier with a semantic version (suggested
format `my-project-name@1.0.0`). Defaults to `NIL`.")
(defvar *dist* nil
   "The distribution of the application. Distributions are used to disambiguate
build or deployment variants of the same release of an application. For
example, the dist can be the build number of an Xcode build or the version
code of an Android build. Defaults to `NIL`.")

(defgeneric dump-event-interface (interface))

(define-json-object (event :ctor %make-event :include envelope-item)
  "[Events](https://develop.sentry.dev/sdk/data-model/envelope-items/#event) are
the fundamental data that clients, often through the use of an SDK, send to the
Sentry server.

Events are packed into `ENVELOPE`s.

See `MAKE-EVENT`"
  (event-id "" :type string)
  (timestamp (local-time:to-rfc3339-timestring (local-time:now)) :type string)
  (platform "other" :type string)
  (level :error :type level :format "\"~(~a~)\"")
  (logger nil :type string :optional t)
  (transaction nil :type string :optional t)
  (server-name *server-name* :type string :optional t)
  (release *release* :type string :optional t)
  (dist *dist* :type string :optional t)
  (tags nil :type list :optional t :format "{~{\"~(~a~)\":~s~^,~}}")
  (modules (get-modules) :type list :format "{~{~s:~s~^,~}}")
  (extra nil :type list :optional t :format "{~{\"~(~a~)\":~a~^,~}}")
  (fingerprint nil :type list :optional t :format "[~{~s~^,~}]")
  (interfaces nil :type list :optional t :raw t :format "~{~s:~a~^,~}"
                  :processor (lambda (interfaces)
                               (reverse-plist
                                (mapcan #'dump-event-interface interfaces))))
  (environment *environment* :type string))

(declaim (inline make-event))
(defun make-event (&key envelope
                        (timestamp nil timestamp-supplied-p)
                        (platform "other")
                        (level :error)
                        logger
                        transaction
                        (server-name *server-name*)
                        (release *release*)
                        (dist *dist*)
                        tags
                        (environment *environment*)
                        (modules (get-modules))
                        extra
                        fingerprint)
  "Creates `EVENT` item in the given `ENVELOPE`.

* `TIMESTAMP` indicates when the event was created. The format is a string as
defined in [RFC 3339](https://tools.ietf.org/html/rfc3339). Defaults to
current time.
* `PLATFORM` is a string representing the platform the client is submitting
from. This will be used by the Sentry interface to customize various
components in the interface. Defaults to `other`.
* `LEVEL` is an optional record severity. Acceptable values are `:FATAL`,
`:ERROR`, `:WARNING`, `:INFO` and `:DEBUG`. Defaults to `:ERROR`.
* `LOGGER` is the optional name of the logger which created the record.
* `TRANSACTION` is the name of the transaction which caused this exception.
For example, in a web app, this might be the route name.
* `SERVER-NAME` identifies the host from which the event was
  recorded. Defaults to the value of `*SERVER-NAME*`.
* `RELEASE` is the optional release version of the application. Defaults to
 the value of `*RELEASE*`.
* `DIST` is the optional distribution of the application. Defaults to the
  value of `*DIST*`.
* `TAGS` is an optional plist of tags for this event. Each tag must be less
  than 200 characters. See `MAKE-TAGS`.
* `ENVIRONMENT` is the environment name, such as `production` or
  `staging`. Defaults to the value of `*ENVIRONMENT*`.
* `MODULES` is an optional plist of modules and their versions. Defaults to
  the result of call to `GET-MODULES`.
* `EXTRA` is an arbitrary mapping of additional metadata to store with the
  event in form of plist. **NOTE** that plist values should be printed with
  `~A` to valid JSON string. See `MAKE-EXTRA`.
* `FINGERPRINT` is an optional list of strings used to dictate the
  deduplication of this event."
  (let ((event (apply #'%make-event
                      :envelope envelope
                      :event-id (envelope-event-id envelope)
                      :platform platform
                      :level level
                      :logger logger
                      :transaction transaction
                      :server-name server-name
                      :release release
                      :dist dist
                      :tags tags
                      :environment environment
                      :modules modules
                      :extra extra
                      :fingerprint fingerprint
                      (when timestamp-supplied-p
                        `(:timestamp ,timestamp)))))
    (push event (envelope-items envelope))
    event))

(declaim (inline make-tags))
(defun make-tags (&rest tags &key &allow-other-keys)
  "Prepares the arbitrary keyword arguments to be used as the `TAGS` for
`MAKE-EVENT`."
  tags)

(declaim (inline make-extra))
(defun make-extra (&rest extra &key &allow-other-keys)
  "Prepares the arbitrary keyword arguments to be used as the `EXTRA` for
`MAKE-EVENT`. **NOTE** that values should be printed with `~A` to valid JSON
string, that means that e.g. string values should be quoted."
  extra)

(declaim (inline add-tags))
(defun add-tags (event &rest tags &key &allow-other-keys)
  "Appends arbitrary keyword arguments as tags on given `EVENT`."
  (setf (event-tags event)
        (nconc (event-tags event) tags))
  event)

(declaim (inline add-extra))
(defun add-extra (event &rest extra &key &allow-other-keys)
  "Appends arbitrary keyword arguments as extra data on given `EVENT`."
  (setf (event-extra event)
        (nconc (event-extra event) extra))
  event)

(defun envelope-event (envelope)
  "Returns `EVENT` item from given `ENVELOPE` or `NIL` if it does not have one."
  (find-if #'event-p (envelope-items envelope)))

(defmethod dump-envelope-item-payload ((item event))
  (babel:string-to-octets
   (escape-nonprintable
    (dump-event item))))
