(in-package #:lisp-sentry)


(defun make-message-event (level control &rest args)
  "Creates an `EVENT` corresponding to simple text message with given `LEVEL`,
using `FORMAT` on given `CONTROL` and `ARGS`."
  (let ((event (make-event :envelope (make-envelope) :level level)))
    (add-sdk event)
    (apply #'add-message event control args)
    (apply #'add-contexts event (make-default-contexts))
    (add-breadcrumbs event)
    event))

(declaim (inline make-condition-event))
(defun make-condition-event (condition &optional
                                       (stack-trace (get-stack-trace))
                                       (thread (bt2:current-thread)))
  "Creates an `EVENT` with `EXCEPTION-INTERFACE` containing information on
`CONDITION`.

See `GET-STACK-TRACE`"
  (let ((event (make-event :envelope (make-envelope)
                           :level (if (typep condition 'warning)
                                      :warning
                                      :error))))
    (add-sdk event)
    (add-exception event condition stack-trace thread)
    (add-debug-meta event)
    (when (> (length (bt2:all-threads)) 1)
      (add-threads event :crashed thread))
    (apply #'add-contexts event (make-default-contexts))
    (add-breadcrumbs event)
    event))

(defun make-user-report-event (comments &key name email)
  "Creates an envelope with dummy `EVENT` item and `USER-REPORT` item. Returns
the event."
  (let ((event (make-message-event :info "User Feedback Event")))
    (make-user-report :envelope (event-envelope event)
                      :email email
                      :name name
                      :comments comments)
    event))

(declaim (type boolean *capture-conditions-outside-dumped-image*))
(defvar *capture-conditions-outside-dumped-image* nil
  "Whether to `CAPTURE` conditions outside of circumstances of dumped Lisp
image, e.g. when running the code via `LOAD`, in interactive REPL session or
Sly/Slime.  Defaults to `NIL`.")

(declaim (inline maybe-capture-condition))
(defun maybe-capture-condition (condition &rest rest)
  "Captures a condition according to settings.

See `*CAPTURE-CONDITIONS-OUTSIDE-DUMPED-IMAGE*`"
  (declare (ignore rest))
  (when (or uiop:*image-dumped-p* *capture-conditions-outside-dumped-image*)
    (capture
     (make-condition-event condition))))

(defmacro make-saved-sources (&key system extra)
  "Specifies the set of source files to be saved for display in stack traces in
Sentry. To be used as `:SOURCES` argument to `WITH-SENTRY` and `INIT!`.

`SYSTEM` specifies whether to save sources of Lisp runtime. Defaults to `NIL`
since it may significantly blow the size of the dumped Lisp image and those
sources are not very interesting usually. Consider using Lisp image
compression when setting this to `T`.

`EXTRA` is arbitrary list of directory pathnames to recursively scan for
`*.lisp` files to be saved."
  `(load-time-value (save-sources
                     :extra-dirs
                     (append ,extra
                             ,(if system
                                  #+sbcl ''("SYS:SRC;" "SYS:CONTRIB;")
                                  #-sbcl nil
                                  nil)))
                    t))

(defun init! (&key
              dsn
              capture-conditions-outside-dumped-image
              variable-elision-length
              project-root
              project-package
              (environment "production")
              server-name
              release
              dist
              (build-type "release")
              app-name
              app-version
              app-build
              startup-time
              (process-before-send #'identity)
              sources)
  "Initializes library settings in a manner similar to `WITH-SENTRY` macro.

See `MAYBE-CAPTURE-CONDITION`"
  (setf *dsn* (or dsn "")
        *capture-conditions-outside-dumped-image*
        capture-conditions-outside-dumped-image
        *variable-elision-length* variable-elision-length
        *project-root* project-root
        *project-package* project-package
        *environment* environment
        *server-name* server-name
        *release* release
        *dist* dist
        *build-type* build-type
        *app-name* app-name
        *app-version* app-version
        *app-build* app-build
        *process-before-send* process-before-send)
  (when startup-time
    (setf *startup-time* startup-time))
  (when sources
    (setf *sources* sources)))

(defmacro with-sentry ((&key
                        (dsn '(or (uiop:getenv "SENTRY_DSN") ""))
                        capture-conditions-outside-dumped-image
                        capture-warnings
                        handled
                        variable-elision-length
                        project-root
                        project-package
                        (environment "production")
                        (server-name '(machine-state:network-info))
                        release
                        dist
                        (build-type "release")
                        app-name
                        app-version
                        app-build
                        (startup-time '(local-time:now))
                        (process-before-send ''identity)
                        (sources '(make-saved-sources)))
                       &body body)
  "Initializes library settings as described below and runs `BODY` wrapped in
`HANDLER-BIND` that `CAPTURE`s conditions popping up. Rule of thumb for placing
this macro in your code — call it as early as possible, but be sure to still
be able to catch all conditions you're interested in.

* `DSN` sets `*DSN*`, a Sentry [Data Source Name](https://docs.sentry.io/concepts/key-terms/dsn-explainer).
Defaults to the value of `SENTRY_DSN` environment variable. If unset, the SDK
will just not send any events.
* `CAPTURE-CONDITIONS-OUTSIDE-DUMPED-IMAGE` sets
  `*CAPTURE-CONDITIONS-OUTSIDE-DUMPED-IMAGE*` which controls whether to
  `CAPTURE` conditions outside of circumstances of dumped Lisp image.
* `CAPTURE-WARNINGS` controls whether `WARNING` conditions are `CAPTURE`d.
* `HANDLED` is form that is evaluated each time the condition is `CAPTURE`d to
  determine the value of `*EXCEPTION-HANDLED*` which indicates whether the
  condition was handled or not.
* `VARIABLE-ELISION-LENGTH` sets `*VARIABLE-ELISION-LENGTH*` which controls
  maximum length of variable values displayed in stack traces.
* `PROJECT-ROOT` sets `*PROJECT-ROOT*` which is absolute path to the project
  directory.
* `PROJECT-PACKAGE` sets `*PROJECT-PACKAGE*` which is name of the project
  package.
* `ENVIRONMENT` sets `*ENVIRONMENT*` which is the environment name, such as
  `production` or `staging`.
* `SERVER-NAME` sets `*SERVER-NAME*` which is the name of the device,
  typically a hostname.
* `RELEASE` sets `*RELEASE*` which is the release version of the application.
* `DIST` sets `*DIST*` which is the distribution of the application.
* `BUILD-TYPE` sets `*BUILD-TYPE*` which is string identifying the kind of
  build.
* `APP-NAME` sets `*APP-NAME*` which is human readable application name, as it
  appears on the platform.
* `APP-VERSION` sets `*APP-VERSION*` which is human readable application
  version, as it appears on the platform.
* `APP-BUILD` sets `*APP-BUILD*` which is internal build identifier, as it
  appears on the platform.
* `STARTUP-TIME` is form that is evaluated to set the value of
  `*STARTUP-TIME*` which is the timestamp when the user started the
  application. If set to `NIL`, `*STARTUP-TIME*` is not set. Defaults
  to `(LOCAL-TIME:NOW)`.
* `PROCESS-BEFORE-SEND` sets `*PROCESS-BEFORE-SEND*` which is the designator
  of function of one argument that processes or cancels the `ENVELOPE` right
  before `CAPTURE` sends it to Sentry.
* `SOURCES` is form that is evaluated to set the value of `*SOURCES*` with the
  source code the Lisp image was built of to display in Sentry stack
  traces. Set to `NIL` to prevent source saving whatsoever. Defaults to the
  call to `MAKE-SAVED-SOURCES`.

See `INIT!`"
  `(progn
     ;; NOTE not using LET here since we want most of those settings in other
     ;; threads as well for the Threads Interface to function correctly
     (setf *dsn* ,dsn
           *capture-conditions-outside-dumped-image*
           ,capture-conditions-outside-dumped-image
           *variable-elision-length* ,variable-elision-length
           *project-root* ,project-root
           *project-package* ,project-package
           *environment* ,environment
           *server-name* ,server-name
           *release* ,release
           *dist* ,dist
           *build-type* ,build-type
           *app-name* ,app-name
           *app-version* ,app-version
           *app-build* ,app-build
           *process-before-send* ,process-before-send
           ,@(when startup-time
               `(*startup-time* ,startup-time))
           ,@(when sources
               `(*sources* ,sources)))
     (flet ((condition-handler (condition)
              (let ((*exception-handled* ,handled))
                (maybe-capture-condition condition))))
       (handler-bind
           ((error #'condition-handler)
            ,@(when capture-warnings `((warning #'condition-handler))))
         ,@body))))
