(in-package #:cl-user)


(defpackage lisp-sentry
  (:documentation "**NOTE: this software is of alpha quality, and the API is
subject to change.**

`lisp-sentry` is a full-featured Common Lisp client library for Sentry,
specifically designed to properly function in dumped Lisp images.

[Sentry](https://sentry.io/about) is application monitoring software which
helps developers diagnose, fix, and optimize the performance of their code.

The key entrypoint of this library is `WITH-SENTRY` macro, which initializes
library settings and runs given body of code within `HANDLER-BIND` which
catches unhandled conditions and sends information on those to Sentry; this is
called *capturing* in Sentry argot.

You can also initialize settings by calling `INIT!` and handle conditions
manually, setting `*EXCEPTION-HANDLED*` and capturing them with
`MAYBE-CAPTURE-CONDITION`.

There's also lower-level API that allows to manually create event envelope by
`MAKE-ENVELOPE`, add the event with `MAKE-CONDITION-EVENT` or even
`MAKE-MESSAGE-EVENT` and capture it by calling `CAPTURE`.

Another extension point of the library is `*PROCESS-BEFORE-SEND*` callback
where you can modify the envelope to be sent, by e.g. adding file attachments
via `MAKE-ATTACHMENT` or user comments via `MAKE-USER-REPORT`.

The most important setting is `*DSN*`, which controls the Sentry instance
that'll receive the reports sent by the library. If not set in code, the
`WITH-SENTRY` macro will try to use `SENTRY_DSN` environment variable. If DSN
is unset, the library will silently discard the captured data.

By default, the library does not capture conditions from interactive
sessions. To do so, set `*CAPTURE-CONDITIONS-OUTSIDE-DUMPED-IMAGE*` to `T`.")
  (:use #:cl)
  (:local-nicknames
   (:machine-state :org.shirakumo.machine-state)
   #+lisp-sentry-gpu (:framebuffers :org.shirakumo.framebuffers))
  (:import-from #:alexandria
                #:array-index #:array-length #:length= #:make-keyword
                #:non-negative-fixnum #:read-file-into-string
                #:read-file-into-byte-vector #:string-designator #:symbolicate
                #:when-let #:when-let* #:with-gensyms)
  (:import-from #:trivial-mimes #:mime)
  ;; contexts/app.lisp
  (:export
   #:*build-type*
   #:*app-name*
   #:*app-version*
   #:*app-build*
   #:*startup-time*
   #:app-context
   #:make-app-context)
  ;; contexts/default.lisp
  (:export
   #:make-default-contexts)
  ;; contexts/device.lisp
  (:export
   #:device-context
   #:make-device-context)
  ;; contexts/feature.lisp
  (:export
   #:feature-flag-context
   #:make-feature-flag-context)
  ;; contexts/gpu.lisp
  (:export
   #:gpu-context
   #:make-gpu-context)
  ;; contexts/os.lisp
  (:export
   #:os-context
   #:make-os-context)
  ;; contexts/runtime.lisp
  (:export
   #:runtime-context
   #:make-runtime-context)
  ;; interfaces/breadcrumbs.lisp
  (:export
   #:breadcrumb
   #:make-breadcrumb-data
   #:make-breadcrumb
   #:breadcrumbs-interface
   #:make-breadcrumbs-interface
   #:add-breadcrumbs)
  ;; interfaces/contexts.lisp
  (:export
   #:contexts-interface
   #:make-contexts-interface
   #:add-contexts
   #:add-context)
  ;; interfaces/debug-meta.lisp
  (:export
   #:debug-meta-image
   #:make-debug-meta-image
   #:debug-meta-interface
   #:make-debug-meta-interface
   #:add-debug-meta)
  ;; interfaces/exception.lisp
  (:export
   #:*exception-handled*
   #:exception-mechanism
   #:make-exception-mechanism
   #:exception-interface
   #:make-exception-interface
   #:add-exception)
  ;; interfaces/message.lisp
  (:export
   #:message-interface
   #:make-message-interface
   #:add-message)
  ;; interfaces/request.lisp
  (:export
   #:request-interface
   #:make-request-interface
   #:add-request)
  ;; interfaces/sdk.lisp
  (:export
   #:sdk-interface
   #:make-sdk-interface
   #:add-sdk)
  ;; interfaces/stack-trace.lisp
  (:export
   #:stack-trace-frame
   #:make-stack-trace-frame
   #:*project-root*
   #:*project-package*
   #:get-stack-trace)
  ;; interfaces/threads.lisp
  (:export
   #:thread
   #:make-thread
   #:threads-interface
   #:make-threads-interface
   #:get-threads
   #:add-threads)
  ;; interfaces/user.lisp
  (:export
   #:user-interface
   #:make-user-interface
   #:add-user)
  ;; items/attachment.lisp
  (:export
   #:attachment
   #:make-attachment)
  ;; items/event.lisp
  (:export
   #:*environment*
   #:*server-name*
   #:*release*
   #:*dist*
   #:event
   #:make-event
   #:make-tags
   #:make-extra
   #:add-tags
   #:add-extra
   #:envelope-event)
  ;; items/user-report.lisp
  (:export
   #:user-report
   #:make-user-report)
  ;; envelope.lisp
  (:export
   #:envelope
   #:make-envelope
   #:*process-before-send*
   #:capture)
  ;; helpers.lisp
  (:export
   #:get-modules)
  ;; lisp-sentry.lisp
  (:export
   #:make-message-event
   #:make-condition-event
   #:make-user-report-event
   #:*capture-conditions-outside-dumped-image*
   #:maybe-capture-condition
   #:make-saved-sources
   #:init!
   #:with-sentry)
  ;; payload.lisp
  (:export
   #:*dsn*)
  ;; pretty-print.lisp
  (:export
   #:*variable-elision-length*)
  ;; source.lisp
  (:export
   #:*sources*))
