(in-package #:lisp-sentry)


(declaim (type (or array-length null) *variable-elision-length*))
(defvar *variable-elision-length* nil
  "If set to non-`NIL`, limits the length of variable values displayed in Sentry
stack traces. Emacs Sly's default is 200 which is reasonable value. Defaults to
`NIL`.")

(defvar *seen-objects*)

(defmacro with-maybe-seen-object (obj &body body)
  `(let* ((seenp (and (boundp '*seen-objects*) (gethash ,obj *seen-objects*)))
          (*seen-objects*
            (if (boundp '*seen-objects*)
                *seen-objects*
                (make-hash-table :test 'eq)))
          (*print-circle* t))
     (if seenp
         (format nil "#~a" seenp)
         (progn
           (setf (gethash ,obj *seen-objects*)
                 (hash-table-count *seen-objects*))
           ,@body))))

(declaim (inline do-pretty-print-object))
(defun do-pretty-print-object (obj)
  (with-maybe-seen-object obj
      (with-output-to-string (stream)
        (let ((*print-readably* nil))
          (print-unreadable-object (obj stream :type t :identity t)
            (format stream "{~{~a ~a~^, ~}}"
                    (mapcan
                     (lambda (slot)
                       (let ((name (closer-mop:slot-definition-name slot)))
                         `(,name
                           ,(if (slot-boundp obj name)
                                (pretty-print (slot-value obj name))
                                "#<UNBOUND>"))))
                     (closer-mop:class-slots (class-of obj)))))))))

(defgeneric do-pretty-print (obj)
  (:method (obj)
    (let ((*print-readably* nil))
      (prin1-to-string obj)))
  (:method ((obj symbol))
    (cond ((null obj) "NIL")
          ((eq obj t) "T")
          (t
           (prin1-to-string obj))))
  (:method ((obj string))
    (prin1-to-string obj))
  (:method ((obj array))
    (with-maybe-seen-object obj
      (with-output-to-string (stream)
        (let ((*print-readably* nil))
          (print-unreadable-object (obj stream :type t :identity t)
            (format stream "#(")
            (loop :with first :of-type boolean := t
                  :for e :across obj
                  :do (unless first
                        (format stream ", "))
                      (format stream "~a" (pretty-print e))
                      (setf first nil))
            (format stream ")"))))))
  (:method ((obj cons))
    (with-maybe-seen-object obj
      (with-output-to-string (stream)
        (format stream "(")
        (loop :for e :on obj
              :for tail := (rest e)
              :for first :of-type boolean := t :then nil
              :do (unless first
                    (format stream " "))
                  (when-let (seen (and (not first)
                                       (gethash e *seen-objects*)))
                    (format stream "#~a" seen)
                    (return))
                  (setf (gethash e *seen-objects*)
                        (hash-table-count *seen-objects*))
                  (format stream "~a" (pretty-print (car e)))
                  (unless (consp tail)
                    (when tail
                      (format stream " . ~a" (pretty-print tail)))
                    (return)))
        (format stream ")"))))
  (:method ((obj hash-table))
    (with-maybe-seen-object obj
      (with-output-to-string (stream)
        (let ((*print-readably* nil))
          (print-unreadable-object (obj stream :type t :identity t)
            (format stream "{")
            (loop :for k :being :the :hash-key :using (hash-value v) :of obj
                  :for first :of-type boolean := t :then nil
                  :do (unless first
                        (format stream ", "))
                      (format stream "~a ~a" (pretty-print k) (pretty-print v)))
            (format stream "}"))))))
  (:method ((obj standard-object))
    (do-pretty-print-object obj))
  (:method ((obj standard-generic-function))
    (with-output-to-string (stream)
      (let ((*print-readably* nil))
        (print-unreadable-object (obj stream :type t :identity t)
          (format stream "~a"
                  (closer-mop:generic-function-name #'print-object))))))
  (:method ((obj condition))
    (do-pretty-print-object obj)))

(defun pretty-print (object)
  (let ((result (do-pretty-print object)))
    (if (and *variable-elision-length* (> (length result) *variable-elision-length*))
        (format nil "~a…" (subseq result 0 *variable-elision-length*))
        result)))
